/*
 * TI BQ257000 charger driver
 *
 * Copyright (c) 2018-2019, Sonos, Inc.
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed "as is" WITHOUT ANY WARRANTY of any
 * kind, whether express or implied; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */


#include <linux/i2c.h>
#include <linux/interrupt.h>
#include <linux/irq.h>
#include <linux/mfd/core.h>
#include <linux/module.h>
#include <linux/regmap.h>
#include <linux/of_device.h>
#include <linux/delay.h>
#include <linux/usb/phy.h>
#include <linux/extcon.h>
#include <linux/delay.h>
#include <linux/power_supply.h>
#include <linux/gpio.h>
#include <linux/of_gpio.h>
#include <linux/slab.h>

enum bc_port_type{
        USB_BC_TYPE_DISCNT = 0,
        USB_BC_TYPE_SDP,
        USB_BC_TYPE_DCP,
        USB_BC_TYPE_CDP,
        USB_BC_TYPE_UNKNOW,
        USB_OTG_POWER_ON,
        USB_OTG_POWER_OFF,
        USB_BC_TYPE_MAX,
};



static int dbg_enable=1;
module_param_named(dbg_level, dbg_enable, int, 0644);

#define DBG(args...) \
	do { \
		if (dbg_enable) { \
			pr_info(args); \
		} \
	} while (0)

#define bq25700_info(fmt, args...) pr_info("bq25700: "fmt, ##args)

#define BQ25700_MANUFACTURER		"Texas Instruments"
#define BQ25700_ID			0x59
#define MAX_CHARGEVOLTAGE		16800

enum bq25700_fields {
	EN_LWPWR, WDTWR_ADJ, IDPM_AUTO_DISABLE,
	EN_OOA, PWM_FREQ, EN_LEARN, IADP_GAIN, IBAT_GAIN,
	EN_LDO, EN_IDPM, CHRG_INHIBIT,
	CHARGE_CURRENT,
	MAX_CHARGE_VOLTAGE,

	AC_STAT, ICO_DONE, IN_VINDPM, IN_IINDPM, IN_FCHRG, IN_PCHRG, IN_OTG,
	F_ACOV, F_BATOC, F_ACOC, SYSOVP_STAT, F_LATCHOFF, F_OTG_OVP, F_OTG_OCP,
	STAT_COMP, STAT_ICRIT, STAT_INOM, STAT_IDCHG, STAT_VSYS, STAT_BAT_REMOV,
	STAT_ADP_REMOV,
	INPUT_CURRENT_DPM,
	OUTPUT_INPUT_VOL, OUTPUT_SYS_POWER,
	OUTPUT_DSG_CUR,	OUTPUT_CHG_CUR,
	OUTPUT_INPUT_CUR, OUTPUT_CMPIN_VOL,
	OUTPUT_SYS_VOL, OUTPUT_BAT_VOL,

	EN_IBAT, EN_PROCHOT_LPWR, EN_PSYS, RSNS_RAC, RSNS_RSR,
	PSYS_RATIO, CMP_REF,	CMP_POL, CMP_DEG, FORCE_LATCHOFF,
	EN_SHIP_DCHG, AUTO_WAKEUP_EN,
	PKPWR_TOVLD_REG, EN_PKPWR_IDPM, EN_PKPWR_VSYS, PKPWER_OVLD_STAT,
	PKPWR_RELAX_STAT, PKPWER_TMAX,	EN_EXTILIM, EN_ICHG_IDCHG, Q2_OCP,
	ACX_OCP, EN_ACOC, ACOC_VTH, EN_BATOC, BATCOC_VTH,
	EN_HIZ, RESET_REG, RESET_VINDPM, EN_OTG, EN_ICO_MODE, BATFETOFF_HIZ,
	PSYS_OTG_IDCHG,
	ILIM2_VTH, ICRIT_DEG, VSYS_VTH, EN_PROCHOT_EXT, PROCHOT_WIDTH,
	PROCHOT_CLEAR, INOM_DEG,
	IDCHG_VTH, IDCHG_DEG, PROCHOT_PROFILE_COMP, PROCHOT_PROFILE_ICRIT,
	PROCHOT_PROFILE_INOM, PROCHOT_PROFILE_IDCHG,
	PROCHOT_PROFILE_VSYS, PROCHOT_PROFILE_BATPRES, PROCHOT_PROFILE_ACOK,
	ADC_CONV, ADC_START, ADC_FULLSCALE, EN_ADC_CMPIN, EN_ADC_VBUS,
	EN_ADC_PSYS, EN_ADC_IIN, EN_ADC_IDCHG, EN_ADC_ICHG, EN_ADC_VSYS,
	EN_ADC_VBAT,

	OTG_VOLTAGE,
	OTG_CURRENT,
	INPUT_VOLTAGE,
	MIN_SYS_VOLTAGE,
	INPUT_CURRENT,

	MANUFACTURE_ID,
	DEVICE_ID,

	F_MAX_FIELDS
};

enum charger_t {
	USB_TYPE_UNKNOWN_CHARGER,
	USB_TYPE_NONE_CHARGER,
	USB_TYPE_USB_CHARGER,
	USB_TYPE_AC_CHARGER,
	USB_TYPE_CDP_CHARGER,
	DC_TYPE_DC_CHARGER,
	DC_TYPE_NONE_CHARGER,
};

enum usb_status_t {
	USB_STATUS_NONE,
	USB_STATUS_USB,
	USB_STATUS_AC,
	USB_STATUS_PD,
	USB_STATUS_OTG,
};

enum tpyec_port_t {
	USB_TYPEC_0,
	USB_TYPEC_1,
};

struct bq25700_init_data {
	u32 ichg;
	u32 max_chg_vol;
	u32 input_voltage;
	u32 input_current;
	u32 sys_min_voltage;
	u32 otg_voltage;
	u32 otg_current;
};

struct bq25700_state {
	u8 ac_stat;
	u8 ico_done;
	u8 in_vindpm;
	u8 in_iindpm;
	u8 in_fchrg;
	u8 in_pchrg;
	u8 in_otg;
	u8 fault_acov;
	u8 fault_batoc;
	u8 fault_acoc;
	u8 sysovp_stat;
	u8 fault_latchoff;
	u8 fault_otg_ovp;
	u8 fault_otg_ocp;
};

struct bq25700_device {
	struct i2c_client		*client;
	struct device			*dev;
	struct power_supply		*supply_charger;
	char				model_name[I2C_NAME_SIZE];
	struct regmap			*regmap;
	struct regmap_field		*rmap_fields[F_MAX_FIELDS];
	int				chip_id;
	struct bq25700_init_data	init_data;
	struct bq25700_state		state;
};

static const struct reg_field bq25700_reg_fields[] = {
	[EN_LWPWR] = REG_FIELD(0x12, 15, 15),
	[WDTWR_ADJ] = REG_FIELD(0x12, 13, 14),
	[IDPM_AUTO_DISABLE] = REG_FIELD(0x12, 12, 12),
	[EN_OOA] = REG_FIELD(0x12, 10, 10),
	[PWM_FREQ] = REG_FIELD(0x12, 9, 9),
	[EN_LEARN] = REG_FIELD(0x12, 5, 5),
	[IADP_GAIN] = REG_FIELD(0x12, 4, 4),
	[IBAT_GAIN] = REG_FIELD(0x12, 3, 3),
	[EN_LDO] = REG_FIELD(0x12, 2, 2),
	[EN_IDPM] = REG_FIELD(0x12, 1, 1),
	[CHRG_INHIBIT] = REG_FIELD(0x12, 0, 0),
	[CHARGE_CURRENT] = REG_FIELD(0x14, 6, 12),
	[MAX_CHARGE_VOLTAGE] = REG_FIELD(0x15, 4, 14),
	[AC_STAT] = REG_FIELD(0x20, 15, 15),
	[ICO_DONE] = REG_FIELD(0x20, 14, 14),
	[IN_VINDPM] = REG_FIELD(0x20, 12, 12),
	[IN_IINDPM] = REG_FIELD(0x20, 11, 11),
	[IN_FCHRG] = REG_FIELD(0x20, 10, 10),
	[IN_PCHRG] = REG_FIELD(0x20, 9, 9),
	[IN_OTG] = REG_FIELD(0x20, 8, 8),
	[F_ACOV] = REG_FIELD(0x20, 7, 7),
	[F_BATOC] = REG_FIELD(0x20, 6, 6),
	[F_ACOC] = REG_FIELD(0x20, 5, 5),
	[SYSOVP_STAT] = REG_FIELD(0x20, 4, 4),
	[F_LATCHOFF] = REG_FIELD(0x20, 2, 2),
	[F_OTG_OVP] = REG_FIELD(0x20, 1, 1),
	[F_OTG_OCP] = REG_FIELD(0x20, 0, 0),
	[STAT_COMP] = REG_FIELD(0x21, 6, 6),
	[STAT_ICRIT] = REG_FIELD(0x21, 5, 5),
	[STAT_INOM] = REG_FIELD(0x21, 4, 4),
	[STAT_IDCHG] = REG_FIELD(0x21, 3, 3),
	[STAT_VSYS] = REG_FIELD(0x21, 2, 2),
	[STAT_BAT_REMOV] = REG_FIELD(0x21, 1, 1),
	[STAT_ADP_REMOV] = REG_FIELD(0x21, 0, 0),
	[INPUT_CURRENT_DPM] = REG_FIELD(0x22, 8, 14),
	[OUTPUT_INPUT_VOL] = REG_FIELD(0x23, 8, 15),
	[OUTPUT_SYS_POWER] = REG_FIELD(0x23, 0, 7),
	[OUTPUT_CHG_CUR] = REG_FIELD(0x24, 8, 14),
	[OUTPUT_DSG_CUR] = REG_FIELD(0x24, 0, 6),
	[OUTPUT_INPUT_CUR] = REG_FIELD(0x25, 8, 15),
	[OUTPUT_CMPIN_VOL] = REG_FIELD(0x25, 0, 7),
	[OUTPUT_SYS_VOL] = REG_FIELD(0x26, 8, 15),
	[OUTPUT_BAT_VOL] = REG_FIELD(0x26, 0, 7),

	[EN_IBAT] = REG_FIELD(0x30, 15, 15),
	[EN_PROCHOT_LPWR] = REG_FIELD(0x30, 13, 14),
	[EN_PSYS] = REG_FIELD(0x30, 12, 12),
	[RSNS_RAC] = REG_FIELD(0x30, 11, 11),
	[RSNS_RSR] = REG_FIELD(0x30, 10, 10),
	[PSYS_RATIO] = REG_FIELD(0x30, 9, 9),
	[CMP_REF] = REG_FIELD(0x30, 7, 7),
	[CMP_POL] = REG_FIELD(0x30, 6, 6),
	[CMP_DEG] = REG_FIELD(0x30, 4, 5),
	[FORCE_LATCHOFF] = REG_FIELD(0x30, 3, 3),
	[EN_SHIP_DCHG] = REG_FIELD(0x30, 1, 1),
	[AUTO_WAKEUP_EN] = REG_FIELD(0x30, 0, 0),
	[PKPWR_TOVLD_REG] = REG_FIELD(0x31, 14, 15),
	[EN_PKPWR_IDPM] = REG_FIELD(0x31, 13, 13),
	[EN_PKPWR_VSYS] = REG_FIELD(0x31, 12, 12),
	[PKPWER_OVLD_STAT] = REG_FIELD(0x31, 11, 11),
	[PKPWR_RELAX_STAT] = REG_FIELD(0x31, 10, 10),
	[PKPWER_TMAX] = REG_FIELD(0x31, 8, 9),
	[EN_EXTILIM] = REG_FIELD(0x31, 7, 7),
	[EN_ICHG_IDCHG] = REG_FIELD(0x31, 6, 6),
	[Q2_OCP] = REG_FIELD(0x31, 5, 5),
	[ACX_OCP] = REG_FIELD(0x31, 4, 4),
	[EN_ACOC] = REG_FIELD(0x31, 3, 3),
	[ACOC_VTH] = REG_FIELD(0x31, 2, 2),
	[EN_BATOC] = REG_FIELD(0x31, 1, 1),
	[BATCOC_VTH] = REG_FIELD(0x31, 0, 0),
	[EN_HIZ] = REG_FIELD(0x32, 15, 15),
	[RESET_REG] = REG_FIELD(0x32, 14, 14),
	[RESET_VINDPM] = REG_FIELD(0x32, 13, 13),
	[EN_OTG] = REG_FIELD(0x32, 12, 12),
	[EN_ICO_MODE] = REG_FIELD(0x32, 11, 11),
	[BATFETOFF_HIZ] = REG_FIELD(0x32, 1, 1),
	[PSYS_OTG_IDCHG] = REG_FIELD(0x32, 0, 0),
	[ILIM2_VTH] = REG_FIELD(0x33, 11, 15),
	[ICRIT_DEG] = REG_FIELD(0x33, 9, 10),
	[VSYS_VTH] = REG_FIELD(0x33, 6, 7),
	[EN_PROCHOT_EXT] = REG_FIELD(0x33, 5, 5),
	[PROCHOT_WIDTH] = REG_FIELD(0x33, 3, 4),
	[PROCHOT_CLEAR] = REG_FIELD(0x33, 2, 2),
	[INOM_DEG] = REG_FIELD(0x33, 1, 1),
	[IDCHG_VTH] = REG_FIELD(0x34, 10, 15),
	[IDCHG_DEG] = REG_FIELD(0x34, 8, 9),
	[PROCHOT_PROFILE_COMP] = REG_FIELD(0x34, 6, 6),
	[PROCHOT_PROFILE_ICRIT] = REG_FIELD(0x34, 5, 5),
	[PROCHOT_PROFILE_INOM] = REG_FIELD(0x34, 4, 4),
	[PROCHOT_PROFILE_IDCHG] = REG_FIELD(0x34, 3, 3),
	[PROCHOT_PROFILE_VSYS] = REG_FIELD(0x34, 2, 2),
	[PROCHOT_PROFILE_BATPRES] = REG_FIELD(0x34, 1, 1),
	[PROCHOT_PROFILE_ACOK] = REG_FIELD(0x34, 0, 0),
	[ADC_CONV] = REG_FIELD(0x35, 15, 15),
	[ADC_START] = REG_FIELD(0x35, 14, 14),
	[ADC_FULLSCALE] = REG_FIELD(0x35, 13, 13),
	[EN_ADC_CMPIN] = REG_FIELD(0x35, 7, 7),
	[EN_ADC_VBUS] = REG_FIELD(0x35, 6, 6),
	[EN_ADC_PSYS] = REG_FIELD(0x35, 5, 5),
	[EN_ADC_IIN] = REG_FIELD(0x35, 4, 4),
	[EN_ADC_IDCHG] = REG_FIELD(0x35, 3, 3),
	[EN_ADC_ICHG] = REG_FIELD(0x35, 2, 2),
	[EN_ADC_VSYS] = REG_FIELD(0x35, 1, 1),
	[EN_ADC_VBAT] = REG_FIELD(0x35, 0, 0),
	[OTG_VOLTAGE] = REG_FIELD(0x3B, 6, 13),
	[OTG_CURRENT] = REG_FIELD(0x3C, 8, 14),
	[INPUT_VOLTAGE] = REG_FIELD(0x3D, 6, 13),
	[MIN_SYS_VOLTAGE] = REG_FIELD(0x3E, 8, 13),
	[INPUT_CURRENT] = REG_FIELD(0x3F, 8, 14),

	[MANUFACTURE_ID] = REG_FIELD(0xFE, 0, 7),
	[DEVICE_ID] = REG_FIELD(0xFF, 0, 7),
};


static const struct regmap_range bq25700_readonly_reg_ranges[] = {
	regmap_reg_range(0x20, 0x26),
	regmap_reg_range(0xFE, 0xFF),
};


static const struct regmap_range bq25700_volatile_reg_ranges[] = {
	regmap_reg_range(0x12, 0x12),
	regmap_reg_range(0x14, 0x15),
	regmap_reg_range(0x20, 0x26),
	regmap_reg_range(0x30, 0x35),
	regmap_reg_range(0x3B, 0x3F),
	regmap_reg_range(0xFE, 0xFF),
};

static const struct regmap_access_table bq25700_volatile_regs = {
	.yes_ranges = bq25700_volatile_reg_ranges,
	.n_yes_ranges = ARRAY_SIZE(bq25700_volatile_reg_ranges),
};

static const struct regmap_config bq25700_regmap_config = {
	.reg_bits = 8,
	.val_bits = 16,

	.max_register = 0xFF,
	.cache_type = REGCACHE_RBTREE,

	.volatile_table = &bq25700_volatile_regs,
	.val_format_endian = REGMAP_ENDIAN_LITTLE,
};


static struct bq25700_device *bq25700_charger;

static int bq25700_field_read(struct bq25700_device *charger,
			      enum bq25700_fields field_id)
{
	int ret=0;
	int val=0;
	ret = regmap_field_read(charger->rmap_fields[field_id], &val);
	if (ret<0)
		return ret;
	return val;
}

int bq25700_field_write(struct bq25700_device *charger,
			       enum bq25700_fields field_id, unsigned int val)
{
	return regmap_field_write(charger->rmap_fields[field_id], val);
}

static int bq25700_get_chip_state(struct bq25700_device *charger,
				  struct bq25700_state *state)
{
	int i, ret;

	struct {
		enum bq25700_fields id;
		u8 *data;
	} state_fields[] = {
		{AC_STAT,	&state->ac_stat},
		{ICO_DONE,	&state->ico_done},
		{IN_VINDPM,	&state->in_vindpm},
		{IN_IINDPM, &state->in_iindpm},
		{IN_FCHRG,	&state->in_fchrg},
		{IN_PCHRG,	&state->in_pchrg},
		{IN_OTG,	&state->in_otg},
		{F_ACOV,	&state->fault_acov},
		{F_BATOC,	&state->fault_batoc},
		{F_ACOC,	&state->fault_acoc},
		{SYSOVP_STAT,	&state->sysovp_stat},
		{F_LATCHOFF,	&state->fault_latchoff},
		{F_OTG_OVP,	&state->fault_otg_ovp},
		{F_OTG_OCP,	&state->fault_otg_ocp},
	};

	for (i = 0; i < ARRAY_SIZE(state_fields); i++) {
		ret = bq25700_field_read(charger, state_fields[i].id);
		if (ret < 0)
			return ret;

		*state_fields[i].data = ret;
	}

	return 0;
}

int bq25700_dump_regs(struct bq25700_device *charger)
{
	u32 val = 0;

	regmap_read(charger->regmap, 0x12, &val);
	DBG("REG0x12 : 0x%x\n", val);
	regmap_read(charger->regmap, 0x14, &val);
	DBG("REG0x14 : 0x%x\n", val);
	regmap_read(charger->regmap, 0x15, &val);
	DBG("REG0x15 : 0x%x\n", val);
	regmap_read(charger->regmap, 0x20, &val);
	DBG("REG0x20 : 0x%x\n", val);
	regmap_read(charger->regmap, 0x21, &val);
	DBG("REG0x21 : 0x%x\n", val);
	regmap_read(charger->regmap, 0x22, &val);
	DBG("REG0x22 : 0x%x\n", val);
	regmap_read(charger->regmap, 0x23, &val);
	DBG("REG0x23 : 0x%x\n", val);
	regmap_read(charger->regmap, 0x24, &val);
	DBG("REG0x24 : 0x%x\n", val);
	regmap_read(charger->regmap, 0x25, &val);
	DBG("REG0x25 : 0x%x\n", val);
	regmap_read(charger->regmap, 0x26, &val);
	DBG("REG0x26 : 0x%x\n", val);
	regmap_read(charger->regmap, 0x30, &val);
        DBG("REG0x30 : 0x%x\n", val);
        regmap_read(charger->regmap, 0x31, &val);
        DBG("REG0x31 : 0x%x\n", val);
        regmap_read(charger->regmap, 0x32, &val);
        DBG("REG0x32 : 0x%x\n", val);
        regmap_read(charger->regmap, 0x33, &val);
        DBG("REG0x33 : 0x%x\n", val);
        regmap_read(charger->regmap, 0x34, &val);
        DBG("REG0x34 : 0x%x\n", val);
        regmap_read(charger->regmap, 0x35, &val);
	DBG("REG0x35 : 0x%x\n", val);
	regmap_read(charger->regmap, 0x3b, &val);
	DBG("REG0x3b : 0x%x\n", val);
	regmap_read(charger->regmap, 0x3c, &val);
	DBG("REG0x3c : 0x%x\n", val);
	regmap_read(charger->regmap, 0x3d, &val);
	DBG("REG0x3d : 0x%x\n", val);
	regmap_read(charger->regmap, 0x3e, &val);
	DBG("REG0x3e : 0x%x\n", val);
	regmap_read(charger->regmap, 0x3f, &val);
	DBG("REG0x3f : 0x%x\n", val);
	regmap_read(charger->regmap, 0xfe, &val);
	DBG("REG0xfe : 0x%x\n", val);
	regmap_read(charger->regmap, 0xff, &val);
	DBG("REG0xff : 0x%x\n", val);
	return 0;
}


int bq25700_init(struct bq25700_device *charger)
{
	struct bq25700_state state;
	int ret;

	charger->init_data.ichg = bq25700_field_read(charger, CHARGE_CURRENT) * 64;

	ret = bq25700_field_read(charger, MAX_CHARGE_VOLTAGE) * 16;

	if (ret<0) {
		charger->init_data.max_chg_vol = MAX_CHARGEVOLTAGE;
	} else {
		charger->init_data.max_chg_vol = ret;
	}
	charger->init_data.sys_min_voltage = 1024 + bq25700_field_read(charger, MIN_SYS_VOLTAGE ) *256 ;
	charger->init_data.input_voltage = 3200 + bq25700_field_read(charger, INPUT_VOLTAGE) * 64;

	charger->init_data.input_current = bq25700_field_read(charger, INPUT_CURRENT) * 50;

	charger->init_data.otg_voltage = 4480 + bq25700_field_read(charger, OTG_VOLTAGE) * 64;
        charger->init_data.otg_current = bq25700_field_read(charger, OTG_CURRENT) * 50;

	DBG("charger->init_data.ichg=%d\n", charger->init_data.ichg);
	DBG("charger->init_data.input_voltage=%d\n", charger->init_data.input_voltage);
	DBG("charger->init_data.input_current=%d\n", charger->init_data.input_current);
	DBG("charger->init_data.max_chg_vol=%d\n", charger->init_data.max_chg_vol);

	bq25700_get_chip_state(charger, &state);

	charger->state = state;

	DBG("status:\n");
        DBG("AC_STAT:  %d\n", state.ac_stat);
        DBG("ICO_DONE: %d\n", state.ico_done);
        DBG("IN_VINDPM: %d\n", state.in_vindpm);
        DBG("IN_IINDPM: %d\n", state.in_iindpm);
        DBG("IN_FCHRG: %d\n", state.in_fchrg);
        DBG("IN_PCHRG: %d\n", state.in_pchrg);
        DBG("IN_OTG: %d\n", state.in_otg);
        DBG("F_ACOV: %d\n", state.fault_acov);
        DBG("F_BATOC: %d\n", state.fault_batoc);
        DBG("F_ACOC: %d\n", state.fault_acoc);
        DBG("SYSOVP_STAT: %d\n", state.sysovp_stat);
        DBG("F_LATCHOFF: %d\n", state.fault_latchoff);
        DBG("F_OTGOVP: %d\n", state.fault_otg_ovp);
        DBG("F_OTGOCP: %d\n", state.fault_otg_ocp);

	return 0;
}

static enum power_supply_property bq25700_power_supply_props[] = {
	POWER_SUPPLY_PROP_MANUFACTURER,
	POWER_SUPPLY_PROP_STATUS,
	POWER_SUPPLY_PROP_ONLINE,
	POWER_SUPPLY_PROP_HEALTH,
	POWER_SUPPLY_PROP_CONSTANT_CHARGE_CURRENT,
	POWER_SUPPLY_PROP_CONSTANT_CHARGE_VOLTAGE,
	POWER_SUPPLY_PROP_CONSTANT_CHARGE_VOLTAGE_MAX,
};

static int bq25700_power_supply_get_property(struct power_supply *psy,
					     enum power_supply_property psp,
					     union power_supply_propval *val)
{
	int ret;
	struct bq25700_device *bq = power_supply_get_drvdata(psy);
	struct bq25700_state state;

	state = bq->state;

	ret = bq25700_field_read(bq, AC_STAT);
        if (ret==1) state.ac_stat = 1;
        else if (ret==0) state.ac_stat = 0;

	switch (psp) {
        case POWER_SUPPLY_PROP_STATUS:
                ret =  bq25700_field_read(bq, IN_FCHRG);
                if ( ret > 0 ) state.in_fchrg = ret;
                ret =  bq25700_field_read(bq, IN_PCHRG);
                if ( ret > 0 ) state.in_pchrg = ret;

                if (!state.ac_stat)
                        val->intval = POWER_SUPPLY_STATUS_DISCHARGING;
                else if (state.in_fchrg == 1 || state.in_pchrg == 1)
                        val->intval = POWER_SUPPLY_STATUS_CHARGING;
                else
                        val->intval = POWER_SUPPLY_STATUS_DISCHARGING;

                break;

        case POWER_SUPPLY_PROP_MANUFACTURER:
                val->strval = BQ25700_MANUFACTURER;
                break;

        case POWER_SUPPLY_PROP_ONLINE:
                val->intval = state.ac_stat;
                break;

	case POWER_SUPPLY_PROP_HEALTH:
                ret =  bq25700_field_read(bq, F_BATOC);
                if ( ret >= 0 ) state.fault_batoc = ret ;
                ret =  bq25700_field_read(bq, F_ACOC);
                if ( ret >= 0 ) state.fault_acoc = ret ;
                ret =  bq25700_field_read(bq, F_ACOV);
                if ( ret >= 0 ) state.fault_acov = ret ;

                if (!state.fault_acoc &&
                    !state.fault_acov && !state.fault_batoc)
                        val->intval = POWER_SUPPLY_HEALTH_GOOD;
                else if (state.fault_batoc)
                        val->intval = POWER_SUPPLY_HEALTH_OVERVOLTAGE;
                break;

        case POWER_SUPPLY_PROP_CONSTANT_CHARGE_CURRENT:
                val->intval = 64 * bq25700_field_read(bq, CHARGE_CURRENT);
                break;

        case POWER_SUPPLY_PROP_CONSTANT_CHARGE_VOLTAGE:
                if (!state.ac_stat) {
                        val->intval = 0;
                        break;
                }

                ret = bq25700_field_read(bq, OUTPUT_BAT_VOL);
                if (ret < 0)
                        val->intval = -1;
                else {
                        val->intval = 2880 + ret * 64;
                }
                break;

	case POWER_SUPPLY_PROP_CONSTANT_CHARGE_VOLTAGE_MAX:
                val->intval = bq->init_data.max_chg_vol;
                break;

        default:
                return -EINVAL;
        }
        return 0;
}

static char *bq25700_charger_supplied_to[] = {
	"charger",
};

static const struct power_supply_desc bq25700_power_supply_desc = {
	.name = "bq25700",
	.type = POWER_SUPPLY_TYPE_USB,
	.properties = bq25700_power_supply_props,
	.num_properties = ARRAY_SIZE(bq25700_power_supply_props),
	.get_property = bq25700_power_supply_get_property,
};

int bq25700_power_supply_init(struct bq25700_device *charger)
{
	struct power_supply_config psy_cfg = { .drv_data = charger, };

	psy_cfg.supplied_to = bq25700_charger_supplied_to;
	psy_cfg.num_supplicants = ARRAY_SIZE(bq25700_charger_supplied_to);

	charger->supply_charger =
		power_supply_register(charger->dev,
				      &bq25700_power_supply_desc,
				      &psy_cfg);

	return PTR_ERR_OR_ZERO(charger->supply_charger);
}

static int bq25700_probe(struct i2c_client *client,
			 const struct i2c_device_id *id)
{
	struct i2c_adapter *adapter = to_i2c_adapter(client->dev.parent);
	struct device *dev = &client->dev;
	struct bq25700_device *charger;
	struct device_node *charger_np;
	int ret = 0;
	u32 i = 0;

	if (!i2c_check_functionality(adapter, I2C_FUNC_SMBUS_WORD_DATA)) {
		DBG("i2c_check_functionality error\n");
		return -EIO;
	}
	charger = devm_kzalloc(&client->dev, sizeof(*charger), GFP_KERNEL);
	if (!charger) {
		DBG("%s:%d dev_kzalloc error\n", __func__, __LINE__);
		return -EINVAL;
	}
	charger->client = client;
	charger->dev = dev;

	charger_np = of_find_compatible_node(NULL, NULL, "ti,bq25700");
	if (charger_np) {
		charger->regmap = devm_regmap_init_i2c(client,
						       &bq25700_regmap_config);
		if (IS_ERR(charger->regmap)) {
			dev_err(&client->dev, "Failed to initialize regmap\n");
			return -EINVAL;
		}

		for (i = 0; i < ARRAY_SIZE(bq25700_reg_fields); i++) {
			const struct reg_field *reg_fields = bq25700_reg_fields;
			charger->rmap_fields[i] =
				devm_regmap_field_alloc(dev,
				charger->regmap,
				reg_fields[i]);
			if (IS_ERR(charger->rmap_fields[i])) {
				dev_err(dev, "cannot allocate regmap field\n");
				return PTR_ERR(charger->rmap_fields[i]);
			}
		}
	} else {
		DBG("no matching device for t1, bq25700\n");
		return -EIO;
	}

	i2c_set_clientdata(client, charger);

	charger->chip_id = bq25700_field_read(charger, DEVICE_ID);

	if (charger->chip_id < 0) {
		dev_err(dev, "Cannot read chip ID.\n");
		return charger->chip_id;
	}

	ret = bq25700_init(charger);
	if (ret < 0) {
		dev_err(dev, "Cannot initialize the chip.\n");
		return ret;
	}


	bq25700_power_supply_init(charger);
	bq25700_charger = charger;
	if (dbg_enable)
		bq25700_dump_regs(charger);

	return ret;
}
static void bq25700_shutdown(struct i2c_client *client)
{
	DBG("%s:%d  \n", __func__, __LINE__);
}
static int bq25700_remove(struct i2c_client *client)
{
	struct bq25700_device *charger = i2c_get_clientdata(client);

        power_supply_unregister(charger->supply_charger);

	DBG("%s:%d  \n", __func__, __LINE__);
	return 0;

}

#ifdef CONFIG_PM_SLEEP
static int bq25700_pm_suspend(struct device *dev)
{
	return 0;
}

static int bq25700_pm_resume(struct device *dev)
{
	return 0;
}
#endif

static SIMPLE_DEV_PM_OPS(bq25700_pm_ops, bq25700_pm_suspend, bq25700_pm_resume);

static const struct i2c_device_id bq25700_i2c_ids[] = {
	{ "bq25700"},
	{ },
};
MODULE_DEVICE_TABLE(i2c, bq25700_i2c_ids);

#ifdef CONFIG_OF
static const struct of_device_id bq25700_of_match[] = {
	{ .compatible = "ti,bq25700", },
	{ },
};
MODULE_DEVICE_TABLE(of, bq25700_of_match);
#else
static const struct of_device_id bq25700_of_match[] = {
	{ },
};
#endif

static struct i2c_driver bq25700_driver = {
	.probe		= bq25700_probe,
	.remove         = bq25700_remove,
	.shutdown	= bq25700_shutdown,
	.id_table	= bq25700_i2c_ids,
	.driver = {
		.name		= "bq25700-charger",
		.pm		= &bq25700_pm_ops,
		.of_match_table	= of_match_ptr(bq25700_of_match),
	},
};
module_i2c_driver(bq25700_driver);

MODULE_DESCRIPTION("TI BQ25700A Driver");
MODULE_LICENSE("GPL v2");
MODULE_AUTHOR ("Sonos, Inc.");
