/*
 * Copyright (c) 2016-2018, Sonos, Inc.
 *
 * SPDX-License-Identifier:     GPL-2.0
 *
 * Driver for generic Sonos microphone control.
 */
#include <linux/version.h>
#include <linux/cdev.h>
#include <linux/module.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>
#include <linux/uaccess.h>

#include "blackbox.h"
#include "sonos_device.h"
#include "micctl.h"
#include "mdp.h"
#include "sonos_lock.h"

static int devno = -1;

int (*micctl_init_funcs[])(void) = MICCTL_HW_INITS;
void (*micctl_exit_funcs[])(void) = MICCTL_HW_EXITS;

struct micctl_data {
	struct adc_ops	*ops;
	uint32_t	ref_count:31;
	uint32_t	mute_state:1;
	struct cdev	chr_dev;
} global_data;

void micctl_register_callbacks(struct adc_ops *ops)
{
	struct adc_ops *ops_ptr = global_data.ops;
	ops->next = NULL;
	if (!global_data.ops) {
		global_data.ops = ops;
		return;
	}

	while (ops_ptr->next) {
		ops_ptr = ops_ptr->next;
	}
	ops_ptr->next = ops;
}

int micctl_is_muted(void)
{
	struct adc_ops *ops_ptr = global_data.ops;
	int ret = 1;
	if (!global_data.ops) {
		bb_log(BB_MOD_MICCTL, BB_LVL_WARNING, "Tried to check mute on the mic hw before any callbacks were registered.");
		return -ENODEV;
	}

	while (ops_ptr) {
		if (!ops_ptr->is_muted) {
			bb_log_dbg(BB_MOD_MICCTL, "%p does not have is_muted.", ops_ptr);
			ops_ptr = ops_ptr->next;
			continue;
		}
		ret &= ops_ptr->is_muted(ops_ptr);
		if (ret) {
			break;
		}
		ops_ptr = ops_ptr->next;
	}
	return ret;
}

int micctl_set_mute(int on)
{
	struct adc_ops *ops_ptr = global_data.ops;
	int ret = 0;
	if (!global_data.ops) {
		bb_log(BB_MOD_MICCTL, BB_LVL_WARNING, "Tried to set mute before any callbacks were registered.");
		return -ENODEV;
	}

	bb_log(BB_MOD_MICCTL, BB_LVL_DEBUG, "Mute %s", (on ? "on" : "off"));

	while (ops_ptr) {
		if (!ops_ptr->mute) {
			bb_log_dbg(BB_MOD_MICCTL, "%p does not have mute.", ops_ptr);
			ops_ptr = ops_ptr->next;
			continue;
		}
		ret = ops_ptr->mute(ops_ptr, on);
		if (ret) {
			break;
		}
		bb_log_dbg(BB_MOD_MICCTL, "Set mute for %p.", ops_ptr);
		ops_ptr = ops_ptr->next;
	}

	if (ret) {
		bb_log(BB_MOD_MICCTL, BB_LVL_ERR, "Unable to set mute (%d).", ret);
	} else {
		global_data.mute_state = on;
	}

	return ret;
}

static int micctl_show(struct seq_file *m, void *v)
{
	struct adc_ops *ops_ptr = global_data.ops;
	int mutes = 0;

	while (ops_ptr) {
		if (ops_ptr->mute) {
			mutes++;
		}
		ops_ptr = ops_ptr->next;
	}
	seq_printf(m, "Sonos Microphone Control\n\n");
	seq_printf(m, "Reference count: %d\n", global_data.ref_count);
	seq_printf(m, "Mute state: %s\n", (global_data.mute_state ? "on" : "off"));
	seq_printf(m, "Mute callbacks registered: %d\n", mutes);

	return 0;
}

static ssize_t micctl_proc_write(struct file *filp, const char __user *buffer, size_t count, loff_t *data)
{
	char buf[200];

	if (!is_mdp_authorized(MDP_AUTH_FLAG_MIC_DBG_ENABLE)) {
		bb_log(BB_MOD_MICCTL, BB_LVL_WARNING, "micctl debug commands are not authorized.");
		return -EACCES;
	}

	if (count >= sizeof(buf)) {
		return -EIO;
	} else if (copy_from_user(buf, buffer, count)) {
		return -EFAULT;
	} else {
		buf[count] = '\0';
	}

	if (strncmp(buf, "mute", 4) == 0) {
		micctl_set_mute(1);
	} else if (strncmp(buf, "unmute", 6) == 0) {
		micctl_set_mute(0);
	}

	return count;
}

const static struct seq_operations micctl_op = {
	.show		= micctl_show
};

static int micctl_proc_open(struct inode *inode, struct file *file)
{
	return single_open(file, micctl_show, PDE_DATA(inode));
}

static struct file_operations micctl_proc_ops = {
	.owner		= THIS_MODULE,
	.open		= micctl_proc_open,
	.write		= micctl_proc_write,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= single_release
};

#define MICCTL_PROCFS_FILE "driver/micctl"

static int micctl_proc_init(void)
{
	struct proc_dir_entry *entry;

	entry = proc_create(MICCTL_PROCFS_FILE, 0666, NULL, &micctl_proc_ops);
	if (!entry) {
		return -EIO;
	}

	return 0;
}

static void micctl_proc_remove(void)
{
	remove_proc_entry(MICCTL_PROCFS_FILE, NULL);
}

static long micctl_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	uint8_t on = 0;
	long ret = 0;
	if (_IOC_DIR(cmd) & _IOC_WRITE) {
		if (copy_from_user(&on, (uint8_t *)arg, sizeof(uint8_t))) {
			return -EACCES;
		}
	}
	switch (_IOC_NR(cmd)) {
	case 0:
	{
		uint32_t ver = MICCTL_VERSION;
		if (copy_to_user((uint32_t *)arg, &ver, sizeof(uint32_t))) {
			return -EACCES;
		}
		break;
	}
	case 1:
	{
		ret = micctl_set_mute((on ? 1 : 0));
		if (ret) {
			bb_log(BB_MOD_MICCTL, BB_LVL_ERR, "MICCTL_SET_MUTE IOCTL failed with %ld.", ret);
		}
		break;
	}
	default:
		bb_log(BB_MOD_MICCTL, BB_LVL_WARNING, "Unrecognized IOCTL %d", _IOC_NR(cmd));
		return -EINVAL;
	}
	return ret;
}


static int micctl_open(struct inode *inodep, struct file *filp)
{
	global_data.ref_count++;
	return 0;
}

static int micctl_release(struct inode *inodep, struct file *filp)
{
	global_data.ref_count--;
	return 0;
}

const struct file_operations micctl_fops = {
	.open		= micctl_open,
	.unlocked_ioctl	= micctl_ioctl,
	.release	= micctl_release,
};

int __init micctl_init(void)
{
	int idx, ret = 0;
	int is_muted;
	struct device *class_dev = NULL;

	devno = MKDEV(MICCTL_MAJOR_NUMBER, 0);
	global_data.ops = NULL;
#ifdef CONFIG_DEVTMPFS
	ret = alloc_chrdev_region(&devno, 0, 1, MICCTL_DEVICE_NAME);
#else
	ret = register_chrdev_region(devno, 1, MICCTL_DEVICE_NAME);
#endif
	if (ret) {
		bb_log(BB_MOD_MICCTL, BB_LVL_ERR, "Couldn't register character device region (%d).", ret);
		return ret;
	}
	cdev_init(&global_data.chr_dev, &micctl_fops);
	global_data.chr_dev.owner = THIS_MODULE;

	ret = cdev_add(&global_data.chr_dev, devno, 1);
	if (ret) {
		bb_log(BB_MOD_MICCTL, BB_LVL_ERR, "Couldn't add character device (%d).", ret);
		unregister_chrdev_region(devno, 1);
		return ret;
	}

	class_dev = sonos_device_create(NULL, devno, NULL, MICCTL_DEVICE_NAME);
	if (IS_ERR(class_dev)) {
		bb_log(BB_MOD_MICCTL, BB_LVL_ERR, "Error creating micctl class.\n");
		cdev_del(&global_data.chr_dev);
		unregister_chrdev_region(devno, 1);
		ret = PTR_ERR(class_dev);
		return ret;
	}

	for(idx = 0; micctl_init_funcs[idx]; idx++) {
		ret = micctl_init_funcs[idx]();
		if (ret == -ENODEV) {
			bb_log(BB_MOD_MICCTL, BB_LVL_INFO, "No mics on this device");
			goto out_err;
		} else if (ret) {
			bb_log(BB_MOD_MICCTL, BB_LVL_ERR, "Failed to init mic device %d (%d)", idx, ret);
			goto out_err;
		}
	}

	micctl_set_mute(1);

	ret = micctl_proc_init();
	if (ret) {
		bb_log(BB_MOD_MICCTL, BB_LVL_ERR, "Failed to init mic procfs (%d).", ret);
	}

	is_muted = micctl_is_muted();
	if (is_muted != -ENODEV) {
		global_data.mute_state = is_muted;
	}

	return ret;
out_err:
	for (; idx >= 0; --idx) {
		if (micctl_exit_funcs[idx]) {
			micctl_exit_funcs[idx]();
		}
	}
	sonos_device_destroy(devno);
	cdev_del(&global_data.chr_dev);
	unregister_chrdev_region(devno, 1);
	return ret;
}
module_init(micctl_init);

void micctl_exit(void)
{
	int idx = 0;
	micctl_proc_remove();
	sonos_device_destroy(devno);
	cdev_del(&global_data.chr_dev);
	unregister_chrdev_region(devno, 1);
	for(; micctl_exit_funcs[idx]; idx++) {
		micctl_exit_funcs[idx]();
	}
}
module_exit(micctl_exit);

MODULE_AUTHOR("Sonos, Inc.");
MODULE_DESCRIPTION("Microphone control driver");
MODULE_LICENSE("GPL");
